//************************************************************************
// Grundkurs Informatik - Kompression durch Differenz-Codierung
//
// H. Ernst, J. Schmidt
// Differenz=0: Code=0000,
// 1: 0001, -1: 1001, 2: 0010, -2: 1010,
// 3: 0011, -3: 1011, 4: 0100, -4: 1100,
// 5: 0101, -5: 1101, 6: 0110, -6: 1110,
// 7: 0111, -7: 1111, |Differenz|>7: 1000
//------------------------------------------------------------------------
#include <stdio.h>
#include <ctype.h>
#include <stdlib.h>
#include <string.h>
#include <math.h>
#define ESC 27
#define BYTE unsigned char
#define MAXLINE 80
#define MAXDATA 800000

// Makros
#define GETH4(x) ((x)>>4) // liefert 4 High-Bits von x
#define GETL4(x) ((x)&15) // liefert 4 Low-Bits von x
#define SETL4(x,y) (((x)&240)|((y)&15)) // kopiert L4 von y in L4 von x
#define SETH4(x,y) (((x)&15) |((y)<<4)) // kopiert L4 von y in H4 von x

// Funktionsprototypen
long diskReadBytes(BYTE text[]);
long diskWriteBytes(long n, BYTE a[]);
long diffCode(long n, BYTE x[], BYTE z[]);
long diffDecode(long nc, BYTE x[], BYTE z[]);
void printCode(long n, BYTE z[]);


//------------------------------------------------------------------------
// Hauptprogramm
//------------------------------------------------------------------------
int main() {
    long i, n = 0, nc = 0;
    BYTE data[MAXDATA];
    BYTE c = 0, z[MAXDATA + MAXDATA / 2];
    char str[2];

    while (c != 'X') {
        printf("\nDifferenz-Codierung - Auswahl (bitte Buchstabe + <enter> drücken)");
        printf("\n===================");
        printf("\nDaten lesen (l) ? ");
        printf("\nCode lesen (r) ? ");
        printf("\nDaten als Text auflisten (t) ? ");
        printf("\nDaten als Zahlen auflisten (z) ? ");
        printf("\nCode auflisten (p) ? ");
        printf("\nDifferenz-Codierung (c) ? ");
        printf("\nDifferenz-Decodierung (d) ? ");
        printf("\nDaten speichern (s) ? ");
        printf("\nCode speichern (w) ? ");
        printf("\nBeenden (x) ? ");
        scanf("%1s", str);

        c = toupper(str[0]);
        if (c == 'L') { // Daten von Platte lesen
            n = diskReadBytes(data);
            if (n > 0)
                printf("Menge der gelesenen Daten: %ld Byte", n);
            else
                printf("\n\nEs wurden keine Daten gelesen!");
        } else if (c == 'R') { // Code von Platte lesen
            nc = diskReadBytes(z);
            if (nc > 0)
                printf("Menge der gelesenen Code-Zeichen: %ld Byte", nc);
            else
                printf("\n\nEs wurde kein Code gelesen!");
        } else if (c == 'T') { // Ausgabe der Daten
            if (n < 1)
                printf("\n\nKeine Daten!\n"); // keine Daten
            else {
                printf("\n\nDaten als Text:\n---------------\n");
                for (i = 0; i < n; i++)
                    printf("%c", data[i]); // als Text
            }
        } else if (c == 'Z') { // Ausgabe der Daten
            if (n < 1)
                printf("\n\nKeine Daten!\n"); // keine Daten
            else {
                printf("\n\nDaten als Zahlen (Bytes):\n");
                printf("-------------------------\n");
                for (i = 0; i < n; i++) // als Zahlen
                    printf("%3d ", data[i]);
            }
        } else if (c == 'P') { // Codierte Daten
            if (nc < 1)
                printf("\n\nKein Code!\n"); // kein Code
            else
                printCode(nc, z); // komprimierte Daten
        } else if (c == 'C') { // Differenz-Code
            nc = diffCode(n, data, z); // Komprimieren
            printf("\n\nKompression durch Differenz-Codierung");
            printf("\nMenge der codierten Daten: %ld Byte", nc);
        } else if (c == 'D') { // Dekomprimieren
            if (nc < 1)
                printf("\n\nKeine Daten!\n"); // keine Daten
            n = diffDecode(nc, data, z); // Dekomprimieren
            printf("\n\nDekompression durch Differenz-Codierung");
            printf("\nMenge der decodierten Daten: %ld Byte", n);
        } else if (c == 'S') { // Code speichern
            if (n > 0)
                n = diskWriteBytes(n, data);
            if (n < 1)
                printf("\nEs wurden keine Daten gespeichert!");
            else
                printf("Menge der gespeicherten Daten: %ld Byte", n);
        } else if (c == 'W') { // Code speichern
            if (nc > 0)
                nc = diskWriteBytes(nc, z);
            if (nc < 1)
                printf("\nEs wurde kein Code gespeichert!");
            else
                printf("Menge der gespeicherten Code-Zeichen: %ld Byte", nc);
        } else if (c != 'X')
            printf("\n\nUnbekannter Befehl!");
        printf("\n\n");
    }
    return 0;
}
//------------------------------------------------------------------------
// Differenz-Codierung der n Byte des Arrays x in das Array z
// Differenz=0: Code=0000,
// 1: 0001, -1: 1001, 2: 0010, -2: 1010,
// 3: 0011, -3: 1011, 4: 0100, -4: 1100,
// 5: 0101, -5: 1101, 6: 0110, -6: 1110,
// 7: 0111, -7: 1111, |Differenz|>7: 1000
// Rückgabewert: Anzahl der Bytes in z
//------------------------------------------------------------------------
long diffCode(long n, BYTE x[], BYTE z[]) {
    long i, j = 1, d, high = 1;
    if (n < 1)
        return 0;
    z[0] = x[0]; // erstes Byte von x in z
    for (i = 1; i < n; i++) { // Daten durchlaufen
        d = x[i] - x[i - 1]; // Differenz
        if (d < 8 && d > -8) { // -8 < Differenz < 8
            if (d < 0)
                d = 8 - d; // Differenz negativ
            if (high) {
                z[j] = SETH4(z[j], d); // d in höherwertige 4 Bit von z
                z[j] = SETL4(z[j], 8); // 8 in niederwertige 4 Bit von z
            } else {
                z[j] = SETL4(z[j], d); // d in niederwertige 4 Bits von z
                j++;
//				z[j] = SETL4(z[j++], d); // d in niederwertige 4 Bits von z
            }
        } else { // d=8: Differenz zu groß
            if (high) {
                z[j] = SETH4(z[j], 8); // 1000 in höherwertige 4 Bits
                z[j] = SETL4(z[j], GETH4(x[i])); // H4 von x in L4 von z, weiter
                j++;
                z[j] = SETH4(z[j], GETL4(x[i])); // L4 von x in H4 von z
                z[j] = SETL4(z[j], 8); // 1000 in L4 von z
            } else {
                z[j] = SETL4(z[j], 8); // 1000 in L4 und weiter
                j++;
//				z[j++] = x[i]; // x in z und weiter
                z[j] = x[i]; // x in z und weiter
                j++;
            }
        }
        high = 1 - high; // High/Low-Umschaltung
    }
    if (!high)
        j++;
    return j;
}
//------------------------------------------------------------------------
// Differenz-Decodierung der nd Bytes des Arrays z in das Array x.
// Rückgabewert: Anzahl der Bytes in x
//------------------------------------------------------------------------
long diffDecode(long nc, BYTE x[], BYTE z[]) {
    long i, n = 1, nc1 = nc - 1;
    BYTE dh, dl; // High und Low 4 Bits
    x[0] = z[0];
    for (i = 1; i < nc; i++) {
        dh = GETH4(z[i]); // H4
        dl = GETL4(z[i]); // L4
        if (dh == 8) { // H4=8: |d| > 7
            if (i >= nc1)
                return n;
            x[n] = SETH4(x[n], dl); // L4 von z in H4 von x
            x[n] = SETL4(x[n], GETH4(z[++i])); // nächstes H4 in L4 von x
            n++;
            dl = GETL4(z[i]); // nächstes L4
            if (dl == 8) {
                if (i >= nc1)
                    return n;
                x[n++] = z[++i]; // L4=8: |d| > 7
            } else { // L4<8: |d| < 8
                if (dl < 8) {
                    x[n] = x[n - 1] + dl; // Differenz > 0
                    n++;
                }
                else {
                    x[n] = x[n - 1] - dl + 8; // Differenz < 0
                    n++;
                }
            }
        } else { // H4 !=8 : d=-7...7
            if (dh < 8) {
                x[n] = x[n - 1] + dh; // Differenz >= 0
                n++;
            }
            else {
                x[n] = x[n - 1] - dh + 8; // Differenz < 0
                n++;
            }
            if (dl == 8) {
                if (i >= nc1)
                    return n;
                x[n] = z[++i]; // L4=0: |d| > 7
                n++;
            } else { // L4 != 8: d=-7...7
                if (dl < 8) {
                    x[n] = x[n - 1] + dl; // Differenz >= 0
                    n++;
                }
                else {
                    x[n] = x[n - 1] - dl + 8; // Differenz < 0
                    n++;
                }
            }
        }
    }
    return n;
}
//------------------------------------------------------------------------
// Differenz-Codierung z ausgeben
//------------------------------------------------------------------------
void printCode(long nc, BYTE z[]) {
    long i, nc1 = nc - 1;
    BYTE x, dh, dl; // Higher und lower 4 Bits
    printf("\n\nKomprimierte Daten:\n-------------------\n");
    printf("HL%d ", z[0]);
    for (i = 1; i < nc; i++) {
        dh = GETH4(z[i]);
        dl = GETL4(z[i]);
        if (dh == 8) { // H4=8: |d| > 7
            if (i < nc1) {
                x = SETH4(x, GETL4(z[i])); // L4 von z in H4 von x
                x = SETL4(x, GETH4(z[++i])); // nächstes H4 von z in L4 von x
                printf("H8 LH%d ", x);
            }
            dl = GETL4(z[i]); // nächstes L4
            printf("L%d ", dl);
            if (i < nc1)
                printf("HL%d ", z[++i]);
        } else { // H4 != 8: |d| < 8
            printf("H%d ", dh);
            printf("L%d ", dl);
        }
    }
}

//------------------------------------------------------------------------
// Datei von Platte lesen
// Rückgabewert:
// -1: Datei nicht gefunden
// >=0: Anzahl der gelesenen Zeichen
//------------------------------------------------------------------------
long diskReadBytes(BYTE data[]) {
    FILE *f;
    char name[MAXLINE];
    long n = 0;
    printf("\n\nDatei von Disk lesen\nDateiname = ? ");
    scanf("%s", name); // Dateinamen einlesen
    if (!(f = fopen(name, "rb"))) { // Datei öffnen
        printf("\nDatei " "%s" " nicht gefunden!\n", name);
        return -1;
    }
    while (n < MAXDATA && !feof(f))
        fscanf(f, "%c", &data[n++]); // lesen
    fclose(f); // Datei schließen
    return n - 1; // Anzahl der gelesenen Zeichen
}
//------------------------------------------------------------------------
// Das Array a mit Dimension n wird in einer Datei gespeichert.
// Rückgabewert:
// -1: Fehler beim Anlegen der Datei
// >=0: Anzahl der gespeicherten Elemente
//------------------------------------------------------------------------
long diskWriteBytes(long n, BYTE a[]) {
    FILE *f;
    long i;
    char name[MAXLINE];
    printf("\n\nDatei auf Disk schreiben\nDateiname = ? ");
    scanf("%s", name); // Dateinamen einlesen
    if ((f = fopen(name, "wb+")) == NULL)
        return -1; // Datei-Fehler
    for (i = 0; i < n; i++)
        fprintf(f, "%c", a[i]); // Datei schreiben
    fclose(f); // Datei schließen
    return i; // Anzahl der Datensätze
}
