/*
 * Grundkurs Informatik - Beispielprogramm zur LZW-Datenkompression
 * J. Schmidt, 2016
*/

#include <assert.h>
#include <stdlib.h>
#include <stdio.h>

#include "bytekette.h"

void initByteKette(t_byte_kette *k)
{
	k->anzZeichen = 0;
	k->maxAnzZeichen = 0;
	k->zeichen = NULL;
	k->speicherBlockGroesse = 100;
}

void speicherAnlegenByteKette(t_byte_kette *k)
{
	k->zeichen = (t_byte *) malloc(k->speicherBlockGroesse * sizeof(t_byte));
	assert(k->zeichen != NULL);

	k->maxAnzZeichen = k->speicherBlockGroesse;
}

void speicherAnlegenFesteGroesseByteKette(t_byte_kette *k, size_t groesse)
{
	k->zeichen = (t_byte *) malloc(groesse * sizeof(t_byte));
	assert(k->zeichen != NULL);

	k->maxAnzZeichen = groesse;
}


void speicherFreigebenByteKette(t_byte_kette *k)
{
	assert(k->zeichen != NULL);
	if(k->zeichen != NULL)
		free(k->zeichen);

	k->maxAnzZeichen = 0;
	k->anzZeichen = 0;
	k->zeichen = NULL;
}

void speicherVergroessernByteKette(t_byte_kette *k)
{
	t_byte *p;
	assert(k->zeichen != NULL);
	p = (t_byte *)realloc(k->zeichen, (k->maxAnzZeichen + k->speicherBlockGroesse) * sizeof(t_byte));
	assert(p != NULL);

	k->zeichen = p;
	k->maxAnzZeichen = k->maxAnzZeichen + k->speicherBlockGroesse;
}

void einzelZeichenAnhaengenByteKette(t_byte_kette *k, const t_byte zeichen)
{
	// pr�fe, ob Speicher ausreicht
	if(k->anzZeichen == k->maxAnzZeichen)
		speicherVergroessernByteKette(k);

	k->zeichen[k->anzZeichen] = zeichen;
	(k->anzZeichen)++;
}

// Setze Zeichenkette a = b
void ersetzenByteKette(t_byte_kette *a, const t_byte_kette b)
{
	size_t i;

	// prüfe, ob Speicher ausreicht
	if(a->maxAnzZeichen != b.maxAnzZeichen)
	{
		speicherFreigebenByteKette(a);
		speicherAnlegenFesteGroesseByteKette(a, b.maxAnzZeichen);
	}

	// alles kopieren
	a->anzZeichen = b.anzZeichen;
	a->speicherBlockGroesse = b.speicherBlockGroesse;
	for(i = 0; i < b.anzZeichen; i++)
		a->zeichen[i] = b.zeichen[i];

}

// Setze Zeichenkette a = b  (b ist ein einzelnes Zeichen)
void ersetzenByteKetteEinzel(t_byte_kette *a, const t_byte b)
{
	// prüfe, ob Speicher ausreicht
	if(a->maxAnzZeichen == 0)
	{
		speicherFreigebenByteKette(a);
		speicherAnlegenByteKette(a);
	}

	// alles kopieren
	a->anzZeichen = 1;
	a->zeichen[0] = b;

}

// liefert true, wenn die beiden Strukturen die gleichen Zeichen enthalten
bool byteKettenIdentisch(const t_byte_kette a, const t_byte_kette b)
{
	bool identisch = false;

	if(a.anzZeichen == b.anzZeichen)
	{
		size_t i;
		identisch = true;

		for(i = 0; i < a.anzZeichen; i++)
			if(a.zeichen[i] != b.zeichen[i])
			{
				identisch = false;
				break;
			}
	}

	return identisch;
}

// printf für eine Zeichenkette
// nur druckbare Zeichen werden ausgegeben, andere durch "-" ersetzt
void printfByteKette(const t_byte_kette a)
{
	size_t i;

	for(i = 0; i < a.anzZeichen; i++)
	{
		if(a.zeichen[i] > 31)
			printf("%c", a.zeichen[i]);
		else
			printf("-");
	}
}

// printf für eine Zeichenkette - rechtsbündig mit Länge laenge
// nur druckbare Zeichen werden ausgegeben, andere durch "-" ersetzt
void printfByteKetteRechts(const t_byte_kette a, unsigned int laenge)
{
	size_t i;
	int start = (int) laenge - (int) a.anzZeichen;
	if(start < 0) start = 0;

	for(i = 0; i < start; i++)
	{
		printf(" ");
	}

	for(i = 0; i < a.anzZeichen; i++)
	{
		if(a.zeichen[i] > 31)
			printf("%c", a.zeichen[i]);
		else
			printf("-");
	}
}

